import tensorflow as tf
import numpy as np

class YOLOLoss(tf.keras.losses.Loss):
  def __init__(self, name='yolo_loss', *args, **kwargs):
    super(YOLOLoss, self).__init__(name = name)

    self.anchor_box   = np.array([[ 7.34375,  7.96875],
                                  [ 6.65625, 10.46875],
                                  [ 9.28125, 10.40625],
                                  [ 5.34375,  9.21875],
                                  [10.28125,  6.65625]])
    self.lambda_coord = 5.0
    self.lambda_noobj = 1.0
    self.threshold    = 0.5
    self.min_max_delta_xy = (-0.1, 0.1)
    self.min_max_delta_wh = (-1.21, 1.21)

    cy, cx = tf.meshgrid(tf.range(13), tf.range(13))

    self.cell = tf.stack([cx, cy], axis = -1)
    self.cell = tf.expand_dims(self.cell, axis = -2)
    self.cell = tf.tile(self.cell, [1, 1, 5, 1])
    self.cell = tf.cast(self.cell, tf.float32)

  def call(self, y_true, y_pred):
    true_conf = y_true[..., 0]
    pred_conf = y_pred[..., 0]
    pred_conf = tf.cast(pred_conf, tf.float32)

    true_coord = y_true[..., 1:5]
    pred_coord = y_pred[..., 1:5]

    true_xy = y_true[..., 1:3]
    pred_xy = y_pred[..., 1:3]

    true_wh = y_true[..., 3:5]
    pred_wh = y_pred[..., 3:5]

    true_prob = y_true[..., 5:]
    pred_prob = y_pred[..., 5:]

    objectness = tf.where(true_conf == 1, 1., 0.)
    objectness = tf.cast(objectness, tf.float32)

    ious = self.iou(true_conf, true_coord, pred_conf, pred_coord)
    ious = tf.cast(ious, tf.float32)

    xy_loss = tf.reduce_sum(tf.square(pred_xy - true_xy), axis = -1)
    xy_loss = tf.cast(xy_loss, tf.float32)
    xy_loss = self.lambda_coord * tf.reduce_sum(objectness * xy_loss)

    wh_loss = tf.reduce_sum(tf.square(tf.sqrt(pred_wh) - tf.sqrt(true_wh)), axis = -1)
    wh_loss = tf.cast(wh_loss, tf.float32)
    wh_loss = self.lambda_coord * tf.reduce_sum(objectness * wh_loss)

    coord_loss = xy_loss + wh_loss

    object_loss     = tf.reduce_sum(objectness * tf.square(pred_conf - ious))
    object_loss     = tf.cast(object_loss, tf.float32)

    no_object_loss  = self.lambda_noobj * tf.reduce_sum((1 - objectness) * tf.square(pred_conf - 0))
    no_object_loss  = tf.cast(no_object_loss, tf.float32)

    class_loss      = tf.reduce_sum(tf.square(pred_prob - true_prob), axis = -1)
    class_loss      = tf.cast(class_loss, tf.float32)
    class_loss      = tf.reduce_sum(objectness * class_loss)

    total_loss = coord_loss + object_loss + no_object_loss + class_loss

    return total_loss

  def convXY(self, delta_xy):
    xy_grid = delta_xy * self.anchor_box + 0.5
    xy      = 32 * xy_grid + 32 * self.cell

    return tf.round(xy)

  def convWH(self, delta_wh):
    wh_grid = self.anchor_box * tf.math.exp(delta_wh)
    wh      = wh_grid * 32

    return tf.round(wh)

  def iou(self, true_conf, true_coord, pred_conf, pred_coord):
    true_delta_xy = true_coord[..., :2]
    pred_delta_xy = pred_coord[..., :2]

    true_delta_wh = true_coord[..., 2:]
    pred_delta_wh = pred_coord[..., 2:]

    true_xy = self.convXY(true_delta_xy)
    true_wh = self.convWH(true_delta_wh)

    pred_xy = self.convXY(pred_delta_xy)
    pred_wh = self.convWH(pred_delta_wh)

    x1, y1 = true_xy[..., 0], true_xy[..., 1]
    w1, h1 = true_wh[..., 0], true_wh[..., 1]

    x2, y2 = pred_xy[..., 0], pred_xy[..., 1]
    w2, h2 = pred_wh[..., 0], pred_wh[..., 1]

    intersection = tf.math.minimum(w1, w2)
    intersection *= tf.math.minimum(h1, h2)

    area1 = w1 * h1
    area2 = w2 * h2

    union = area1 + area2 - intersection

    iou = intersection / union

    return iou

  def normalize_delta(self, val, min_val, max_val):
    return (val - min_val) / (max_val - min_val)

  def denormalize_delta(self, val, min_val, max_val):
    return val * (max_val - min_val) + min_val